@@ -30,7 +30,11 @@ int fcntl(int fd, int cmd, ...)
 		}
 		ret = __syscall(SYS_fcntl, fd, F_DUPFD_CLOEXEC, 0);
 		if (ret != -EINVAL) {
+#ifdef __EMSCRIPTEN__
+			if (ret >= 0) __wasi_fd_close(ret);
+#else
 			if (ret >= 0) __syscall(SYS_close, ret);
+#endif
 			return __syscall_ret(-EINVAL);
 		}
 		ret = __syscall(SYS_fcntl, fd, F_DUPFD, arg);
@@ -39,7 +43,6 @@ int fcntl(int fd, int cmd, ...)
 	}
 	switch (cmd) {
 	case F_SETLK:
-	case F_SETLKW:
 	case F_GETLK:
 	case F_GETOWN_EX:
 	case F_SETOWN_EX:
